/**
 * Data class. An instance contains the reference set data.
 * @author Zhihao
 *
 */
class Instance {

	double[] supplies;			//levels of supply from supply nodes. length is <number of supply nodes>
	
	double[] demandLBs;			//hard lower bound on demands. length is <number of demand nodes>
	double[] demandUBs;			//hard upper bound on demands. length is <number of demand nodes>
	double totalDemandUB = 0;	//sum of demand upper bounds
	
	double[][] scenarios;		//demand samples in the instance. dimensions are <number of scenarios>*<number of demand nodes>
	
	double[] mu;				//mean. length is <number of demand nodes>
	double[] sigmasq;			//variance. length is <number of demand nodes>
	
	Instance(double[] supplies_,double[] demandLBs_,double[] demandUBs_,double[][] scenarios_) {
		supplies = supplies_;
		demandLBs = demandLBs_;
		demandUBs = demandUBs_;
		for(int i = 0; i < this.demandUBs.length; i++) {
			totalDemandUB += demandUBs[i];
		}
		scenarios = scenarios_;
		mu = new double[scenarios[0].length];
		sigmasq = new double[scenarios[0].length];
		for(int i = 0; i < mu.length; i++) {
			for(int s = 0; s < scenarios.length; s++) {
				mu[i] += scenarios[s][i];
			}
			mu[i] /= scenarios.length;
			for(int s = 0; s < scenarios.length; s++) {
				sigmasq[i] += Math.pow(scenarios[s][i]-mu[i],2);
			}
			sigmasq[i] /= scenarios.length;
		}
	}
}
