import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;
import java.util.Random;

import org.apache.commons.io.FileUtils;

import gurobi.GRB;
import gurobi.GRBEnv;
import gurobi.GRBException;

/**
 * Main class. All methods that are to be run should be placed here.
 * @author Zhihao, Hideaki
 *
 */
public class Main {
	
	//These are multiplied to change the scales
	static double penaltyFactor = 1;
	static double capCostFactor = 1;
	static double flowCostFactor = 1;

	static SimpleDateFormat sdf = new SimpleDateFormat("HH:mm:ss");

	static double drndpTol = 0.02;	//PROPORTION difference for comparison
	static double L = 0;
	
	static String mainDirectory = "";
	
	static String[] mainDirNames = {"nobel-us"
			//,"nobel-eu"
			//,"pioro40"
			//,"grid 3x3"
			//,"grid 4x4"
			//,"grid 5x5"
			//,"grid 6x6"
			//,"grid 7x7"
	};
	
	static String[] words = {"gamma"
			,"uniform"
	};
	static int[] sizes = {50
			//,100
			//,500
			,1000
	};
	
	//cardinality: multiplied by the number of demand
	static double[] cardinality = {0.0
			,0.4
			//,0.7
			//,1.0
	};
	static int sampleSize = 20; //number of samples extracted from reference
	static double regularDemandParam = 0.075;
	static double lowDemandParam = 0.7;
	

	static double supplyValue = 20;
	// currently, this code only supports uniform LB,UB,numIntervals
	static double demandLB = 0.0;
	static double demandUB = 50.0;
	static int numIntervals = 25;
	static double intervalLength = (demandUB-demandLB) / (double)numIntervals;
	
	//reference parameters
	static int numref = 10000;
	static double aveCapCost = 40;
	static double sigCapCost = 6;
	static double aveFlowCost = 10;
	static double sigFlowCost = 2;
	//static double PenaltyCost = 10;
	static double shape = 4;
	static double scale = 5;
	//
	
	
	public static void main(String[] args) {

		String filepath = new java.io.File("").getAbsolutePath();
		String origDir = filepath + "/datafiles_and_results";
		NewFolder.create(origDir);
		
		//5.1-5.3
		NewFolder.create(origDir + "/Experiment1");
		mainDirectory = origDir + "/Experiment1/Instance0";
		NewFolder.create(mainDirectory);
		int numSupply = 3;
		int numDemand = 3;
		makeGraph(numSupply, numDemand);
		runIt();
		makeSummary(mainDirectory + "/summarized means.csv");
		
		//5.4
		//repeat with same setting as Experiment0 except with different supply/demand location
		int numRepeat = 3;
		for(int i=1; i<numRepeat; i++) {
			mainDirectory = origDir + "/Experiment1/Instance" + i;
			NewFolder.create(mainDirectory);
			repeatGraph(origDir + "/Experiment1/Instance0",1);
			runIt();
			makeSummary(mainDirectory + "/summarized means.csv");
		}
		
		//5.5.1
		//change the unit penalty cost to 10, 20, 50 times the default penalty
		double[] penalties = {10,20,50};
		NewFolder.create(origDir + "/Experiment2");
		for (int i=1; i<=penalties.length; i++) {
			penaltyFactor = penalties[i-1];
			mainDirectory = origDir + "/Experiment2/Instance"+i;
			NewFolder.create(mainDirectory);
			repeatGraph(origDir + "/Experiment1/Instance0",2);
			runIt();
			makeSummary(mainDirectory + "/summarized means.csv");
		}
		penaltyFactor = 1;
		
		//5.5.2
		//change the number of intervals to 5 and 10
		int[] intervals = {10,5};
		NewFolder.create(origDir + "/Experiment3");
		for (int i=1; i<=intervals.length; i++) {
			numIntervals  = intervals[i-1];
			mainDirectory = origDir + "/Experiment3/Instance"+i;
			NewFolder.create(mainDirectory);
			repeatGraph(origDir + "/Experiment1/Instance0",2);
			runIt();
			makeSummary(mainDirectory + "/summarized means.csv");
		}
		numIntervals = 25;
		
		//5.5.3
		//change the number of demand nodes from 2 to 5, changing the value of shape in Gamma distribution accordingly
		int[] numDemands = {2,4,5};
		numSupply = 3;
		NewFolder.create(origDir + "/Experiment4");
		for (int i=1; i<=numDemands.length; i++) {
			numDemand = numDemands[i-1];
			cardinality = new double[] {2.0 / (double)numDemand};
			shape = 12/numDemand;
			mainDirectory = origDir + "/Experiment4/Instance"+i;
			NewFolder.create(mainDirectory);
			makeGraph(numSupply, numDemand);
			runIt();
			makeSummary(mainDirectory + "/summarized means.csv");
		}
		
		
	}
	public static void makeGraph(int numSupply, int numDemand) {
		for(int i = 0; i < mainDirNames.length; i++) {
			String mainDirName = mainDirNames[i];
			makeDirectory(mainDirName);
			switch(mainDirName){
			case "nobel-us":
				SampleMaker.networkParser("graphs/nobel-us--D-B-E-N-C-A-N-N/nobel-us.txt", mainDirectory+"/nobel-us/input files", numSupply, numDemand);
				break;
			case "nobel-eu":
				SampleMaker.networkParser("graphs/nobel-eu--D-B-E-N-C-A-N-N/nobel-eu.txt", mainDirectory+"/nobel-eu/input files", numSupply, numDemand);
				break;
			case "pioro40":
				SampleMaker.networkParser("graphs/pioro40--U-U-M-N-C-A-N-N/pioro40.txt", mainDirectory+"/pioro40/input files", numSupply, numDemand);
				break;
			case "grid 3x3":
				SampleMaker.gridNetwork(mainDirectory+"/grid 3x3/input files", 3, numSupply, numDemand);
				break;
			case "grid 4x4":
				SampleMaker.gridNetwork(mainDirectory+"/grid 4x4/input files", 4, numSupply, numDemand);
				break;
			case "grid 5x5":
				SampleMaker.gridNetwork(mainDirectory+"/grid 5x5/input files", 5, numSupply, numDemand);
				break;
			case "grid 6x6":
				SampleMaker.gridNetwork(mainDirectory+"/grid 6x6/input files", 6, numSupply, numDemand);
				break;
			case "grid 7x7":
				SampleMaker.gridNetwork(mainDirectory+"/grid 7x7/input files", 7, numSupply, numDemand);
				break;
			default:
				throw new java.lang.Error("Support for "+mainDirName+" is lost or does not exist");
			}
			SampleMaker.makeSamples(numSupply,mainDirName);
				
		}
	}
	public static void repeatGraph(String sourceDirName, int mode) {
		/* 
		 * sourceDirName should contain all folders of mainDirNames
		 */
		for(int i = 0; i < mainDirNames.length; i++) {
			String mainDirName = mainDirNames[i];
			makeDirectory(mainDirName);
			switch(mode) {
			case 1:
				//when you change the location of supply/demand but use same data
				SampleMaker.changeLocation(sourceDirName, mainDirName);
				SampleMaker.copySamples(sourceDirName,mainDirName);
				break;
			case 2:
				//when you use identical data
				SampleMaker.copyInstance(sourceDirName, mainDirName);
				break;
			}
		}
	}
	
	
	public static void makeDirectory(String mainDirName) {
		NewFolder.create(mainDirectory + "/" + mainDirName);
		NewFolder.create(mainDirectory + "/" + mainDirName + "/input files");

		NewFolder.create(mainDirectory + "/" + mainDirName + "/input files/sample 1");
		NewFolder.create(mainDirectory + "/" + mainDirName + "/input files/sample 2");
		NewFolder.create(mainDirectory + "/" + mainDirName + "/input files/sample 3");
	}
	public static void runIt() {
		try {		
			GRBEnv env = new GRBEnv();
			env.set(GRB.IntParam.OutputFlag,0);
			//env.set(GRB.IntParam.Presolve,0);
			String mainDirName;
			for(int i = 0; i < mainDirNames.length; i++) {
				mainDirName = mainDirNames[i];
				getSolutions(mainDirName,env);
				getPerformance(mainDirName,env);
			}
		} catch (GRBException e) {
			System.err.println("Gurobi exception: " + e);
		}
	}
	public static void getSolutions(String mainDirName,GRBEnv env) throws GRBException {
		System.out.println("Now getting solutions for "+mainDirName+" (Time: "+sdf.format(new Date())+")");
		String mainDir = mainDirectory + "/"+mainDirName;
		String inputDir = mainDir+"/input files";
		Network network = ReaderWriter.file2network(inputDir+"/network.csv");
		String outputDir = mainDir+"/output files";
		NewFolder.create(outputDir);
		String inputSampleDir;
		String outputSampleDir;
		String modelDir;
		Instance instance;
		Output output;
		for(int i = 1; i <= 3; i++) {
			inputSampleDir = inputDir+"/sample "+i;
			outputSampleDir = outputDir+"/sample "+i;
			NewFolder.create(outputSampleDir);
			
			modelDir = outputSampleDir+"/drndp";
			System.out.println("drndp"+" (Time: "+sdf.format(new Date())+")");
			NewFolder.create(modelDir);
			instance = ReaderWriter.file2instance(network,inputSampleDir+"/drndp.csv");
			DRNDP drndp = new DRNDP(env,network,instance,L);
			output = drndp.solve();
			output.printDRNDP(modelDir);

			for(int j = 0; j < words.length; j++) {
				for(int k = 0; k < sizes.length; k++) {
					modelDir = outputSampleDir+"/"+words[j]+sizes[k];
					System.out.println(words[j]+sizes[k]+" (Time: "+sdf.format(new Date())+")");
					NewFolder.create(modelDir);
					instance = ReaderWriter.file2instance(network,inputSampleDir+"/"+words[j]+sizes[k]+".csv");
					SNDP sndp = new SNDP(env,network,instance);
					output = sndp.solve();
					output.printSNDP(modelDir);
				}
			}
			//added by Hideaki
			for(int j = 0; j < cardinality.length; j++) {
				modelDir = outputSampleDir+"/rndp_"+cardinality[j];
				System.out.println("rndp_"+cardinality[j]+" (Time: "+sdf.format(new Date())+")");
				NewFolder.create(modelDir);
				instance = ReaderWriter.file2instance(network,inputSampleDir+"/drndp.csv");//use instance of drndp
				RNDP rndp = new RNDP(env,network,instance,cardinality[j]);
				output = rndp.solve();
				output.printRNDP(modelDir);
			}
			//
		}
	}
	public static void getPerformance(String mainDirName,GRBEnv env) throws GRBException {
		System.out.println("Now getting performance for "+mainDirName+" (Time: "+sdf.format(new Date())+")");
		String mainDir = mainDirectory + "/"+mainDirName;
		String inputDir = mainDir+"/input files";
		Network network = ReaderWriter.file2network(inputDir+"/network.csv");
		Instance refset = ReaderWriter.file2instance(network,inputDir+"/reference set.csv");
		NDP ndp = new NDP(env,network,refset.supplies);
		String outputDir = mainDir+"/output files";
		String outputSampleDir;
		String modelDir;
		Output output;
		for(int i = 1; i <= 3; i++) {
			outputSampleDir = outputDir+"/sample "+i;
			
			modelDir = outputSampleDir+"/drndp";
			System.out.println("drndp"+" (Time: "+sdf.format(new Date())+")");
			output = getPerformanceHelper(modelDir,ndp,refset);
			output.printPerformance(modelDir);

			for(int j = 0; j < words.length; j++) {
				for(int k = 0; k < sizes.length; k++) {
					modelDir = outputSampleDir+"/"+words[j]+sizes[k];
					System.out.println(words[j]+sizes[k]+" (Time: "+sdf.format(new Date())+")");
					output = getPerformanceHelper(modelDir,ndp,refset);
					output.printPerformance(modelDir);
				}
			}
			
			//added by Hideaki
			for(int j = 0; j < cardinality.length; j++) {
				modelDir = outputSampleDir+"/rndp_"+cardinality[j];
				System.out.println("rndp_"+cardinality[j]+" (Time: "+sdf.format(new Date())+")");
				output = getPerformanceHelper(modelDir,ndp,refset);
				output.printPerformance(modelDir);
			}
			//
		}
	}
	private static Output getPerformanceHelper(String modelDir,NDP ndp,Instance refset) throws GRBException {
		Output output = new Output();
		output.yQuant = new double[refset.scenarios.length][];
		output.tQuant = new double[refset.scenarios.length][];
		output.flowCostQuant = new double[refset.scenarios.length];
		output.penaltyCostQuant = new double[refset.scenarios.length];
		output.x = output.readXfromSummary(modelDir+"/summary.csv");
		for(int ij = 0; ij < ndp.network.arcs.length; ij++) {
			output.capCost += ndp.network.arcs[ij].capCost*output.x[ij];
		}
		ndp.updateX(output.x);
		for(int s = 0; s < refset.scenarios.length; s++) {
			ndp.updateDemands(refset.scenarios[s]);
			ndp.model.optimize();
			//process solution
			output.yQuant[s] = new double[ndp.y.length];
			for(int ij = 0; ij < ndp.y.length; ij++) {
				output.yQuant[s][ij] = ndp.y[ij].get(GRB.DoubleAttr.X);
			}
			output.tQuant[s] = new double[ndp.t.length];
			for(int j = 0; j < ndp.t.length; j++) {
				output.tQuant[s][j] = ndp.t[j].get(GRB.DoubleAttr.X);
			}
			for(int ij = 0; ij < ndp.network.arcs.length; ij++) {
				output.flowCostQuant[s] += ndp.network.arcs[ij].flowCost*output.yQuant[s][ij];
			}
			for(int j = 0; j < ndp.network.demandNodes.length; j++) {
				output.penaltyCostQuant[s] += ndp.network.nodes[ndp.network.demandNodes[j]].penalty*output.tQuant[s][j];
			}
		}
		return output;
	}
	public static void makeSummary(String outputFile) {
		try {
			PrintWriter writer = new PrintWriter(outputFile);
			writer.println(
					"network"
					+",sample"
					+",model"
					+",solve time"
					+",h solve time"
					+",# cuts"
					+",cap cost"
					+",model opt val"
					+",model 2nd stage"
					+",e opt val"
					+",e 2nd stage"
					+",e flow cost"
					+",e penalty cost"
					+",e unmet demand"
					+",unmet demand 75%"
					+",unmet demand 95%"
					+",unmet demand 99%"
					);
			BufferedReader br;
			String mainDir;
			String outputDir;
			String outputSampleDir;
			String modelDir;
			long solveTime;
			long htime;
			int numCuts;
			double capCost;
			double modelOptVal;
			double modelSecondStage;
			double eOptVal;
			double eSecondStage;
			double eFlowCost;
			double[] unmetDemand;
			List<Double> unmetDemandList;
			double ePenaltyCost;
			double eUnmetDemand;
			String str;
			String[] splitstr;
			int counter;
			double temp;
			for(int i = 0; i < mainDirNames.length; i++) {
				mainDir = mainDirectory + "/"+mainDirNames[i];
				outputDir = mainDir+"/output files";
				for(int m = 1; m <= 3; m++) {
					outputSampleDir = outputDir+"/sample "+m;
					
					modelDir = outputSampleDir+"/drndp";
					br = new BufferedReader(new FileReader(modelDir+"/summary.csv"));
					br.readLine();
					solveTime = Long.parseLong(br.readLine());
					modelOptVal = Double.parseDouble(br.readLine());
					capCost = Double.parseDouble(br.readLine());
					modelSecondStage = Double.parseDouble(br.readLine());
					numCuts = Integer.parseInt(br.readLine());
					htime = Long.parseLong(br.readLine());
					br.close();
					br = new BufferedReader(new FileReader(modelDir+"/performance summary.csv"));
					br.readLine();
					eOptVal = Double.parseDouble(br.readLine());
					br.readLine();
					eFlowCost = Double.parseDouble(br.readLine());
					ePenaltyCost = Double.parseDouble(br.readLine());
					eSecondStage = eFlowCost+ePenaltyCost;
					br.close();
					br = new BufferedReader(new FileReader(modelDir+"/tquant.csv"));
					str = br.readLine();
					eUnmetDemand = 0;
					unmetDemandList = new ArrayList<Double>();
					counter = 0;
					while(str != null) {
						counter++;
						splitstr = str.split(",");
						temp = 0;
						for(int n = 0; n < splitstr.length; n++) {
							temp += Double.parseDouble(splitstr[n]);
						}
						eUnmetDemand += temp;
						unmetDemandList.add(temp);
						str = br.readLine();
					}
					eUnmetDemand /= counter;
					unmetDemand = new double[unmetDemandList.size()];
					for(int n = 0; n < unmetDemand.length; n++) {
						unmetDemand[n] = unmetDemandList.get(n);
					}
					Sort.quicksort(unmetDemand);
					br.close();
					writer.println(
							mainDirNames[i]
							+","+m
							+","+"drndp"
							+","+solveTime
							+","+htime
							+","+numCuts
							+","+capCost
							+","+modelOptVal
							+","+modelSecondStage
							+","+eOptVal
							+","+eSecondStage
							+","+eFlowCost
							+","+ePenaltyCost
							+","+eUnmetDemand
							+","+unmetDemand[(int)(unmetDemand.length*0.75)]
							+","+unmetDemand[(int)(unmetDemand.length*0.95)]
							+","+unmetDemand[(int)(unmetDemand.length*0.99)]
							);
					
					for(int j = 0; j < words.length; j++) {
						for(int k = 0; k < sizes.length; k++) {
							modelDir = outputSampleDir+"/"+words[j]+sizes[k];
							br = new BufferedReader(new FileReader(modelDir+"/summary.csv"));
							br.readLine();
							solveTime = Long.parseLong(br.readLine());
							modelOptVal = Double.parseDouble(br.readLine());
							capCost = Double.parseDouble(br.readLine());
							modelSecondStage = Double.parseDouble(br.readLine());
							numCuts = Integer.parseInt(br.readLine());
							htime = Long.parseLong(br.readLine());
							br.close();
							br = new BufferedReader(new FileReader(modelDir+"/performance summary.csv"));
							br.readLine();
							eOptVal = Double.parseDouble(br.readLine());
							br.readLine();
							eFlowCost = Double.parseDouble(br.readLine());
							ePenaltyCost = Double.parseDouble(br.readLine());
							eSecondStage = eFlowCost+ePenaltyCost;
							br.close();
							br = new BufferedReader(new FileReader(modelDir+"/tquant.csv"));
							str = br.readLine();
							eUnmetDemand = 0;
							unmetDemandList = new ArrayList<Double>();
							counter = 0;
							while(str != null) {
								counter++;
								splitstr = str.split(",");
								temp = 0;
								for(int n = 0; n < splitstr.length; n++) {
									temp += Double.parseDouble(splitstr[n]);
								}
								eUnmetDemand += temp;
								unmetDemandList.add(temp);
								str = br.readLine();
							}
							eUnmetDemand /= counter;
							unmetDemand = new double[unmetDemandList.size()];
							for(int n = 0; n < unmetDemand.length; n++) {
								unmetDemand[n] = unmetDemandList.get(n);
							}
							Sort.quicksort(unmetDemand);
							br.close();
							writer.println(
									mainDirNames[i]
									+","+m
									+","+words[j]+sizes[k]
									+","+solveTime
									+","+htime
									+","+numCuts
									+","+capCost
									+","+modelOptVal
									+","+modelSecondStage
									+","+eOptVal
									+","+eSecondStage
									+","+eFlowCost
									+","+ePenaltyCost
									+","+eUnmetDemand
									+","+unmetDemand[(int)(unmetDemand.length*0.75)]
									+","+unmetDemand[(int)(unmetDemand.length*0.95)]
									+","+unmetDemand[(int)(unmetDemand.length*0.99)]
									);
						}
					}
					//added by Hideaki
					for(int j = 0; j < cardinality.length; j++) {
						modelDir = outputSampleDir+"/rndp_"+cardinality[j];
						br = new BufferedReader(new FileReader(modelDir+"/summary.csv"));
						br.readLine();
						solveTime = Long.parseLong(br.readLine());
						modelOptVal = Double.parseDouble(br.readLine());
						capCost = Double.parseDouble(br.readLine());
						modelSecondStage = Double.parseDouble(br.readLine());
						numCuts = Integer.parseInt(br.readLine());
						htime = Long.parseLong(br.readLine());
						br.close();
						br = new BufferedReader(new FileReader(modelDir+"/performance summary.csv"));
						br.readLine();
						eOptVal = Double.parseDouble(br.readLine());
						br.readLine();
						eFlowCost = Double.parseDouble(br.readLine());
						ePenaltyCost = Double.parseDouble(br.readLine());
						eSecondStage = eFlowCost+ePenaltyCost;
						br.close();
						br = new BufferedReader(new FileReader(modelDir+"/tquant.csv"));
						str = br.readLine();
						eUnmetDemand = 0;
						unmetDemandList = new ArrayList<Double>();
						counter = 0;
						while(str != null) {
							counter++;
							splitstr = str.split(",");
							temp = 0;
							for(int n = 0; n < splitstr.length; n++) {
								temp += Double.parseDouble(splitstr[n]);
							}
							eUnmetDemand += temp;
							unmetDemandList.add(temp);
							str = br.readLine();
						}
						eUnmetDemand /= counter;
						unmetDemand = new double[unmetDemandList.size()];
						for(int n = 0; n < unmetDemand.length; n++) {
							unmetDemand[n] = unmetDemandList.get(n);
						}
						Sort.quicksort(unmetDemand);
						br.close();
						writer.println(
								mainDirNames[i]
								+","+m
								+","+"rndp_"+cardinality[j]
								+","+solveTime
								+","+htime
								+","+numCuts
								+","+capCost
								+","+modelOptVal
								+","+modelSecondStage
								+","+eOptVal
								+","+eSecondStage
								+","+eFlowCost
								+","+ePenaltyCost
								+","+eUnmetDemand
								+","+unmetDemand[(int)(unmetDemand.length*0.75)]
								+","+unmetDemand[(int)(unmetDemand.length*0.95)]
								+","+unmetDemand[(int)(unmetDemand.length*0.99)]
								);
					}
					//
				}
			}
			writer.close();
		} catch (IOException e) {
			System.err.println("I/O exception: " + e);
		}
	}
}
