import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Static helper class. Randomly selects some supply and demand nodes and creates the sub-network and sub-instance associated with the subset of nodes
 * @author Zhihao
 *
 */
public class NetworkSizeReducer {
	/**
	 * Reads a network and a network instance (reference set) and reduces it to have exactly 3 randomly selected supply nodes and 3 randomly selected
	 * demande nodes. Generates the sub-network and sub-instance associated with these 3 supply and 3 demand nodes
	 */
	static void reduce() {
		String mainDirName;
		for(int m = 0; m < Main.mainDirNames.length; m++) {
			mainDirName = Main.mainDirNames[m];
			String inputDir =Main.mainDirectory + "/"+mainDirName+"/old input files";
			String newInputDir =Main.mainDirectory + "/"+mainDirName+"/input files";
			NewFolder.create(newInputDir);
			Network network = ReaderWriter.file2network(inputDir+"/network.csv");
			Instance refset = ReaderWriter.file2instance(network,inputDir+"/reference set.csv");
			
			//shuffle 3 random indices each for supply nodes and demand nodes
			List<Integer> supplyNumbers = new ArrayList<Integer>();
			for(int i = 0; i < network.supplyNodes.length; i++) {
				supplyNumbers.add(i);
			}
			Collections.shuffle(supplyNumbers);
			int[] supplySet = new int[3];
			for(int i = 0; i < 3; i++) {
				supplySet[i] = supplyNumbers.get(i);
			}
			Sort.quicksort(supplySet);
			List<Integer> demandNumbers = new ArrayList<Integer>();
			for(int i = 0; i < network.demandNodes.length; i++) {
				demandNumbers.add(i);
			}
			Collections.shuffle(demandNumbers);
			int[] demandSet = new int[3];
			for(int i = 0; i < 3; i++) {
				demandSet[i] = demandNumbers.get(i);
			}
			Sort.quicksort(demandSet);
			/*for(int i = 0; i < 3; i++) {
				System.out.printf(supplySet[i]+"	");
			}
			System.out.printf("\n");
			for(int i = 0; i < 3; i++) {
				System.out.printf(demandSet[i]+"	");
			}
			System.out.printf("\n");*/
			
			//generate new network and reference set with the 3 random supply and demand nodes
			ReaderWriter.network2file(make3SupplyDemandNetwork(supplySet,demandSet,network),newInputDir+"/network.csv");
			ReaderWriter.instance2file(make3SupplyDemandInstance(supplySet,demandSet,refset),newInputDir+"/reference set.csv");

			//generate new samples
			String sampleDir;
			String newSampleDir;
			Instance instance;
			for(int i = 1; i <= 3; i++) {
				sampleDir = inputDir+"/sample "+i;
				newSampleDir = newInputDir+"/sample "+i;
				NewFolder.create(newSampleDir);
				
				instance = ReaderWriter.file2instance(network,sampleDir+"/drndp.csv");
				ReaderWriter.instance2file(make3SupplyDemandInstance(supplySet,demandSet,instance),newSampleDir+"/drndp.csv");

				for(int j = 0; j < Main.words.length; j++) {
					for(int k = 0; k < Main.sizes.length; k++) {
						instance = ReaderWriter.file2instance(network,sampleDir+"/"+Main.words[j]+Main.sizes[k]+".csv");
						ReaderWriter.instance2file(make3SupplyDemandInstance(supplySet,demandSet,instance),newSampleDir+"/"+Main.words[j]+Main.sizes[k]+".csv");
					}
				}
			}
		}
	}
	/**
	 * Creates the sub-network with the given supply node and demand node indices
	 */
	public static Network make3SupplyDemandNetwork(int[] supplySet,int[] demandSet,Network network) {
		String[] nodeNames = new String[network.nodes.length];
		for(int i = 0; i < network.nodes.length; i++) {
			nodeNames[i] = network.nodes[i].name;
		}
		int[] supplyNodes = new int[3];
		int[] demandNodes = new int[3];
		double[] penalties = new double[3];
		for(int i = 0; i < 3; i++) {
			supplyNodes[i] = network.supplyNodes[supplySet[i]];
			demandNodes[i] = network.demandNodes[demandSet[i]];
			penalties[i] = network.nodes[network.demandNodes[demandSet[i]]].penalty;
		}
		int[][] arcIndices = new int[network.arcs.length][2];
		double[] capCosts = new double[network.arcs.length];
		double[] flowCosts = new double[network.arcs.length];
		for(int ij = 0; ij < network.arcs.length; ij++) {
			arcIndices[ij][0] = network.arcs[ij].origin;
			arcIndices[ij][1] = network.arcs[ij].destination;
			capCosts[ij] = network.arcs[ij].capCost;
			flowCosts[ij] = network.arcs[ij].flowCost;
		}
		return new Network(nodeNames,supplyNodes,demandNodes,penalties,arcIndices,capCosts,flowCosts);
	}
	/**
	 * Creates the sub-instance with the given supply node and demand node indices
	 */
	public static Instance make3SupplyDemandInstance(int[] supplySet,int[] demandSet,Instance instance) {
		double[] supplies = new double[3];
		double[] demandLBs = new double[3];
		double[] demandUBs = new double[3];
		double[][] scenarios = new double[instance.scenarios.length][3];
		for(int i = 0; i < 3; i++) {
			supplies[i] = instance.supplies[supplySet[i]];
			demandLBs[i] = instance.demandLBs[demandSet[i]];
			demandUBs[i] = instance.demandUBs[demandSet[i]];
			for(int s = 0; s < instance.scenarios.length; s++) {
				scenarios[s][i] = instance.scenarios[s][demandSet[i]];
			}
		}
		return new Instance(supplies,demandLBs,demandUBs,scenarios);
	}
	/**
	 * Creates the sub-network with the given supply node and demand node indices
	 */
	public static Network make5SupplyDemandNetwork(int[] supplySet,int[] demandSet,Network network) {
		String[] nodeNames = new String[network.nodes.length];
		for(int i = 0; i < network.nodes.length; i++) {
			nodeNames[i] = network.nodes[i].name;
		}
		int[] supplyNodes = new int[5];
		int[] demandNodes = new int[5];
		double[] penalties = new double[5];
		for(int i = 0; i < 5; i++) {
			supplyNodes[i] = network.supplyNodes[supplySet[i]];
			demandNodes[i] = network.demandNodes[demandSet[i]];
			penalties[i] = network.nodes[network.demandNodes[demandSet[i]]].penalty;
		}
		int[][] arcIndices = new int[network.arcs.length][2];
		double[] capCosts = new double[network.arcs.length];
		double[] flowCosts = new double[network.arcs.length];
		for(int ij = 0; ij < network.arcs.length; ij++) {
			arcIndices[ij][0] = network.arcs[ij].origin;
			arcIndices[ij][1] = network.arcs[ij].destination;
			capCosts[ij] = network.arcs[ij].capCost;
			flowCosts[ij] = network.arcs[ij].flowCost;
		}
		return new Network(nodeNames,supplyNodes,demandNodes,penalties,arcIndices,capCosts,flowCosts);
	}
	/**
	 * Creates the sub-instance with the given supply node and demand node indices
	 */
	public static Instance make5SupplyDemandInstance(int[] supplySet,int[] demandSet,Instance instance) {
		double[] supplies = new double[5];
		double[] demandLBs = new double[5];
		double[] demandUBs = new double[5];
		double[][] scenarios = new double[instance.scenarios.length][3];
		for(int i = 0; i < 5; i++) {
			supplies[i] = instance.supplies[supplySet[i]];
			demandLBs[i] = instance.demandLBs[demandSet[i]];
			demandUBs[i] = instance.demandUBs[demandSet[i]];
			for(int s = 0; s < instance.scenarios.length; s++) {
				scenarios[s][i] = instance.scenarios[s][demandSet[i]];
			}
		}
		return new Instance(supplies,demandLBs,demandUBs,scenarios);
	}
}
